import os
import time
import logging
from datetime import datetime
from typing import Union

from pyathena import connect
from pyathena.cursor import Cursor

from cheapodb.utils import CheapoDBException, create_cheapodb_role, normalize_table_name, create_session

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(name)s %(levelname)-8s %(message)s',
    datefmt='%a, %d %b %Y %H:%M:%S'
)

log = logging.getLogger(__name__)


class Database(object):
    """
    A database object represents the components that make up a database in AWS Glue.

    Provides methods for Glue, Athena and S3
    """
    def __init__(self, name: str, description: str = None, auto_create=False, results_prefix='results/',
                 create_iam_role=False, iam_role_arn=None, **kwargs):
        """
        Initialize a database object

        :param name: name of the database
        :param description: optional description of the database (shown in Glue catalog)
        :param auto_create: if True, automatically create the database if it doesn't already exist
        :param results_prefix: optional S3 prefix to store query results
        :param create_iam_role: True to automatically create an IAM role with necessary permissions. One of
        create_iam_role or iam_role_arn is required
        :param iam_role_arn: an optional ARN of an existing IAM role. One of create_iam_role or iam_role_arn is
        required
        :param kwargs:
        """
        self.name = name
        self.description = description
        self.auto_create = auto_create
        self.results_prefix = results_prefix
        self.iam_role_arn = iam_role_arn

        self.session = create_session(
            region_name=kwargs.pop('aws_default_region', os.getenv('AWS_DEFAULT_REGION')),
            aws_access_key_id=kwargs.pop('aws_access_key_id', os.getenv('AWS_ACCESS_KEY_ID')),
            aws_secret_access_key=kwargs.pop('aws_secret_access_key', os.getenv('AWS_SECRET_ACCESS_KEY'))
        )
        self.s3 = self.session.client('s3')
        self.bucket = self.session.resource('s3').Bucket(self.name)
        self.glue = self.session.client('glue')

        if create_iam_role and not self.iam_role_arn:
            self.iam_role_arn = create_cheapodb_role(
                name=f'{self.name}-CheapoDBExecutionRole',
                client=self.session.client('iam'),
                bucket=self.name
            )

        elif not create_iam_role and not self.iam_role_arn:
            msg = f'No IAM role ARN provided and create_iam_role was False. ' \
                  f'Provide either an existing role ARN or set create_iam_role to True.'
            raise CheapoDBException(msg)

        if self.auto_create:
            self.create()

    def create(self):
        """
        Create the database, which consists of an S3 bucket and Glue database object

        :return:
        """
        log.info(f'Creating database {self.name} in {self.session.region_name}')

        bucket_params = dict(
            CreateBucketConfiguration=dict(
                LocationConstraint=self.session.region_name
            )
        )
        response = self.bucket.create(**bucket_params)
        log.debug(response)

        db_params = dict(Name=self.name)
        if self.description:
            db_params['Description'] = self.description
        self.glue.create_database(
            DatabaseInput=db_params
        )
        log.debug(response)

        return

    @property
    def tables(self) -> list:
        """
        Get a listing of all tables that exist in the database

        :return:
        """
        tables = list()
        while True:
            payload = dict(
                DatabaseName=self.name,
                MaxResults=100
            )
            response = self.glue.get_tables(**payload)
            log.debug(response)

            if not response['TableList']:
                break
            if response['NextToken']:
                payload['NextToken'] = response['NextToken']
            tables += response['TableList']

        return tables

    def query(self, sql: str, results_path: str = None) -> dict:
        """
        Execute a query

        :param sql: the Athena-compliant SQL to execute
        :param results_path: optional S3 path to write results. Defaults to current DB bucket and
        instance results_prefix
        :return:
        """
        if not results_path:
            results_path = f's3://{self.name}/{self.results_prefix}'

        cursor = self.export(sql, results_path)
        columns = [column[0] for column in cursor.description]
        log.info(cursor.description)
        for row in cursor:
            yield dict(zip(columns, row))

    def export(self, sql: str, results_path: str) -> Cursor:
        """
        Execute a query and return the cursor.

        Useful for building the athena query results as a file export to a destination bucket/prefix

        :param sql: the Athena-compliant SQL to execute
        :param results_path: required S3 path to write export
        :return: a pyathena cursor object
        """
        cursor = connect(
            s3_staging_dir=results_path,
            region_name=self.session.region_name,
            role_arn=self.iam_role_arn
        ).cursor()
        cursor.execute(sql)
        return cursor

    def create_crawler(self, name, schedule: str = None, table_prefix: str = None, target: str = None,
                       update_behavior='UPDATE_IN_DATABASE', delete_behavior='DELETE_FROM_DATABASE',
                       description: str = None) -> str:
        """
        Create a new Glue crawler.

        Either loads a crawler or creates it if it doesn't exist. The provided name of the crawler
        corresponds to the prefix of the DB bucket it will target.

        :param name: the DB bucket prefix to crawl
        :param schedule: an optional schedule in cron syntax to run the crawler
        :param table_prefix: an optional prefix to apply to the created tables
        :param target: an optional target path to crawl
        :param update_behavior: how the crawler should handle schema updates
        :param delete_behavior: how the crawler should handle deletions
        :param description: Optional crawler description
        :return: the name of the created crawler
        """
        log.info(f'Creating crawler {name}')
        if not target:
            target = f'{self.name}/{name}/'
        if not description:
            description = f'Crawler created by CheapoDB on {datetime.now():%Y-%m-%d %H:%M:%S}'
        try:
            payload = dict(
                Name=name,
                Role=self.iam_role_arn,
                DatabaseName=self.name,
                Description=description,
                Targets=dict(
                    S3Targets=[
                        {
                            'Path': target
                        }
                    ]
                ),
                TablePrefix=f'{normalize_table_name(table_prefix)}_',
                SchemaChangePolicy=dict(
                    UpdateBehavior=update_behavior,
                    DeleteBehavior=delete_behavior
                )
            )
            if schedule:
                payload['Schedule'] = schedule
            self.glue.create_crawler(**payload)
            log.info(f'Created crawler {name}')
            return self.glue.get_crawler(Name=name)['Crawler']['Name']
        except self.glue.exceptions.AlreadyExistsException:
            log.warning(f'Crawler {name} already exists')
            return self.glue.get_crawler(Name=name)['Crawler']['Name']

    def delete_crawler(self, name: str) -> None:
        """
        Delete a Glue crawler

        :param name: the name of the crawler to delete
        :return:
        """
        log.info(f'Deleting crawler {name}')
        self.glue.delete_crawler(
            Name=name
        )
        return

    def update_tables(self, crawler, wait: Union[bool, int] = 60) -> None:
        """
        Run the provided crawler to update the tables in the prefix.
        Optionally wait for completion, which is the default.

        :param crawler: the name of the Glue Crawler to execute
        :param wait: optionally wait for the crawler to complete by providing the number of seconds to wait between
        polling requests. False if this method should return immediately after starting the crawler.
        :return:
        """
        log.info(f'Updating tables with crawler {crawler}')
        response = self.glue.start_crawler(
            Name=crawler
        )
        log.debug(response)
        if wait:
            log.info(f'Waiting for table update to complete...')
            while True:
                response = self.glue.get_crawler(Name=crawler)
                elapsed = response['Crawler']['CrawlElapsedTime'] / 1000
                if response['Crawler']['State'] == 'RUNNING':
                    log.debug(response)
                    log.info(f'Crawler in RUNNING state. Elapsed time: {elapsed} secs')
                    time.sleep(wait)
                    continue
                elif response['Crawler']['State'] == 'STOPPING':
                    log.debug(response)
                    log.info(f'Crawler in STOPPING state')
                    time.sleep(wait)
                    continue
                else:
                    status = response['Crawler']['LastCrawl']['Status']
                    log.debug(response)
                    log.info(f'Crawler in READY state. Table update {status}')
                    break

        return



import os
import logging
from urllib.parse import urlencode
from typing import List

from dask.dataframe import DataFrame

from cheapodb.database import Database
from cheapodb.utils import normalize_table_name


logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s %(name)s %(levelname)-8s %(message)s',
    datefmt='%a, %d %b %Y %H:%M:%S'
)

log = logging.getLogger(__name__)


class Table(object):
    """
    A Table object represents the components that make up a table in AWS Glue.

    Provides methods for Glue, Athena and S3
    """
    def __init__(self, db: Database, name: str, prefix: str):
        """
        Create a Table instance

        :param db: the Database that will contain the Table
        :param name: the name of the Table
        :param prefix: the prefix in the Database where the Table data will reside
        """
        self.db = db
        self.name = normalize_table_name(name)
        self.prefix = prefix

    @property
    def columns(self) -> List[dict]:
        """
        Get a list of table columns

        :return: list of dicts describing the table columns
        """
        return self.describe()['Table']['StorageDescriptor']['Columns']

    def get_versions(self) -> List[dict]:
        """
        Get a list of table versions

        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glue.html#Glue.Client.get_table_versions

        :return:
        """
        versions = list()
        while True:
            payload = dict(
                DatabaseName=self.db.name,
                TableName=self.name,
                MaxResults=100
            )
            response = self.db.glue.get_table_versions(**payload)
            if not response['TableVersions']:
                break
            if response['NextToken']:
                payload['NextToken'] = response['NextToken']
            versions += response['TableVersions']

        return versions

    def describe(self) -> dict:
        """
        Get a reference to the table with metadata.

        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/glue.html#Glue.Client.get_table

        :return: the Glue get_table response
        """
        return self.db.glue.get_table(
            DatabaseName=self.db.name,
            Name=self.name
        )

    def upload(self, f, tags: dict = None) -> None:
        """
        Upload the table data file(s) to the database bucket and prefix

        :param f: path to the file to upload
        :param tags: an optional dict of key:value tags to apply to the uploaded object
        :return:
        """
        target = os.path.join(self.prefix, self.name, self.name)
        log.info(f'Uploading file {f} to {target}')

        extra_args = None
        if tags:
            extra_args = dict(
                Tagging=urlencode(tags)
            )

        self.db.bucket.upload_file(f, target, ExtraArgs=extra_args)
        return

    def download(self, f) -> None:
        """
        Download the originally uploaded table data file to the specified path

        :param f: path where downloaded file will be stored
        :return:
        """
        target = os.path.join(self.prefix, self.name, self.name)
        log.info(f'Downloading file {target} to {f}')
        self.db.bucket.download_file(target, f)
        return

    def as_parquet(self, df: DataFrame, **kwargs) -> None:
        """
        Load a Dask DataFrame as parquet to the database bucket and prefix

        :param df: the Dask DataFrame object to load as parquet
        :param kwargs: additional keyword arguments provided to Dask DataFrame.to_parquet
        :return:
        """
        target = f's3://{os.path.join(self.db.name, self.prefix, self.name)}'
        df.to_parquet(path=target, **kwargs)
        return

    def as_geojson(self, df: DataFrame, compression=None, **kwargs) -> None:
        """

        :param df:
        :param compression:
        :param kwargs:
        :return:
        """
        # import geojson
        # from geomet.wkt import dumps

        target = f's3://{os.path.join(self.db.name, self.prefix, self.name)}'
        df.to_json(target, compression=compression, **kwargs)
        return

    def delete_table(self, include_data: bool = True) -> None:
        """
        Delete a table in a Glue database

        :param include_data: True to include the underlying data in S3
        :return:
        """
        if include_data:
            d = f'{self.prefix}/{self.name}'
            log.info(f'Deleting data at {d}')
            try:
                versions = self.db.s3.list_object_versions(
                    Bucket=self.db.name,
                    Prefix=d
                )['Versions']
                actions = [self.db.s3.delete_object(
                    Bucket=self.db.name,
                    Key=x['Key'],
                    VersionId=x['VersionId']
                ) for x in versions]
                log.debug(actions)
                log.info(f'Deleted data at {d}')
            except KeyError:
                log.warning(f'Data does not exist at {self.db.name}/{d}')
        try:
            log.info(f'Deleting table {self.prefix}_{self.name}')
            response = self.db.glue.delete_table(
                DatabaseName=self.db.name,
                Name=f'{self.prefix}-{self.name}'
            )
            log.debug(response)
            log.info(f'Deleted table {self.prefix}_{self.name}')
        except self.db.glue.exceptions.EntityNotFoundException:
            log.warning(f'Table does not exist in {self.db.name}')
        return

    def from_dataframe(self, df: DataFrame, **kwargs) -> None:
        """
        Load a Dask DataFrame as parquet to the database bucket and prefix

        :param df: the Dask DataFrame object to load as parquet
        :param kwargs: additional keyword arguments provided to Dask DataFrame.to_parquet
        :return:
        """
        target = f's3://{os.path.join(self.db.name, self.prefix, self.name)}'
        df.to_parquet(path=target, engine='fastparquet', **kwargs)
        return



import json
import time
from itertools import islice, chain
from typing import Union, Generator, List

import boto3
from joblib import Parallel, delayed


class Stream(object):
    def __init__(self, name: str):
        self.name = name
        self.firehose = boto3.client('firehose')

    def initialize(self, config):
        if self.exists:
            return
        config = json.loads(config)
        response = self.firehose.create_delivery_stream(**config)
        while True:
            if self.exists:
                break
            time.sleep(10)

        return response

    @property
    def describe(self):
        return self.firehose.describe_delivery_stream(
            DeliveryStreamName=self.name
        )

    @property
    def exists(self) -> bool:
        try:
            self.describe()
            return True
        except self.firehose.exceptions.ResourceNotFoundException:
            return False

    @staticmethod
    def _chunks(iterable: Union[Generator, list], size: int):
        iterator = iter(iterable)
        for first in iterator:
            yield chain([first], islice(iterator, size - 1))

    def from_records(self, records: Union[Generator, List[dict]], threads: int = 4) -> None:
        """
        Ingest from a generator or list of dicts

        :param records: a generator or list of dicts
        :param threads: number of threads for batch putting
        :return:
        """
        Parallel(n_jobs=threads, prefer='threads')(delayed(self.firehose.put_record_batch)(
            DeliveryStreamName=self.name,
            Records=[{'Data': json.dumps(x).encode()} for x in chunk]
        ) for chunk in self._chunks(records, size=500))

        return



import os
import json
import logging
from datetime import datetime
import boto3

log = logging.getLogger(__name__)


class CheapoDBException(Exception):
    pass


def create_session(**kwargs):
    return boto3.session.Session(
        region_name=kwargs.get('aws_default_region', os.getenv('AWS_DEFAULT_REGION')),
        aws_access_key_id=kwargs.get('aws_access_key_id', os.getenv('AWS_ACCESS_KEY_ID')),
        aws_secret_access_key=kwargs.get('aws_secret_access_key', os.getenv('AWS_SECRET_ACCESS_KEY'))
    )


def normalize_table_name(name):
    """Check if the table name is obviously invalid."""
    if not isinstance(name, str):
        raise ValueError()
    name = name.replace('-', '_').strip()
    if not len(name):
        raise ValueError(f'Invalid table name: {name}')
    return name


def create_cheapodb_role(name, client, bucket) -> str:
    """
    Create an AWS IAM service role with the appropriate permissions for Glue and the database's S3 bucket.

    :param name:
    :param client:
    :param bucket:
    :return:
    """
    try:
        response = client.create_role(
            RoleName=name,
            Path='/service-role/',
            Description=f'IAM role created by CheapoDB on {datetime.now():%Y-%m-%d %H:%M:%S}',
            AssumeRolePolicyDocument=json.dumps(dict(
                Version='2012-10-17',
                Statement=[
                    {
                        'Sid': '',
                        'Effect': 'Allow',
                        'Principal': {
                            'Service': 'glue.amazonaws.com'
                        },
                        'Action': 'sts:AssumeRole'
                    }
                ]
            ))
        )
        iam_role_arn = response['Role']['Arn']

        response = client.attach_role_policy(
            RoleName=name,
            PolicyArn='arn:aws:iam::aws:policy/service-role/AWSGlueServiceRole'
        )
        log.debug(response)

        response = client.put_role_policy(
            RoleName=name,
            PolicyName='CheapoDBRolePolicy',
            PolicyDocument=json.dumps(dict(
                Version='2012-10-17',
                Statement=[
                    {
                        'Effect': 'Allow',
                        'Action': [
                            's3:GetObject',
                            's3:PutObject'
                        ],
                        'Resource': [
                            f'arn:aws:s3:::{bucket}*'
                        ]
                    }
                ]
            ))
        )
        log.debug(response)
    except client.exceptions.EntityAlreadyExistsException:
        msg = f'Role already exists for database: CheapoDBRole-{bucket}. ' \
              f'Provide the role ARN as iam_role_arn.'
        raise CheapoDBException(msg)

    return iam_role_arn



import os
import sys

from setuptools import setup, find_packages
from setuptools.command.install import install

VERSION = '0.0.7'


class VerifyVersionCommand(install):
    description = 'verify that git tag matches VERSION prior to publishing to pypi'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = 'Git tag: {0} does not match the version of this app: {1}'.format(
                tag, VERSION
            )
            sys.exit(info)


def read(fname):
    with open(fname) as fp:
        content = fp.read()
    return content


install_requires = [
    'boto3==1.9.180',
    'dask[dataframe]==2.0.0',
    'fastparquet==0.3.1',
    'joblib==0.13.2',
    'pandas==0.24.2',
    'PyAthena==1.6.1',
    's3fs==0.2.1'
]

setup(
    name='cheapodb',
    version=VERSION,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    install_requires=install_requires,
    extras_require={
        'pandas': ['pandas>=0.24.0']
    },
    url='https://github.com/mineralzen/cheapodb',
    keywords=['data', 'aws'],
    packages=find_packages(exclude=('test*',)),
    package_dir={'directaccess': 'directaccess'},
    license='MIT',
    author='Cole Howard',
    author_email='cole@mineralzen.com',
    description='Opinionated implementation of AWS Glue',
    cmdclass={
        'verify': VerifyVersionCommand,
    },
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
)
